"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BtcTransaction = void 0;
const constants_1 = require("../../utils/constants");
const errors_1 = require("../../utils/errors");
const utils_1 = require("../../utils/utils");
const BtcAddress_1 = require("../addressObjects/BtcAddress");
const UtxoTransaction_1 = require("./UtxoTransaction");
class BtcTransaction extends UtxoTransaction_1.UtxoTransaction {
    // Btc specific transaction
    get currencyName() {
        return constants_1.BTC_NATIVE_TOKEN_NAME;
    }
    isValidPkscript(index) {
        const vout = this.extractVoutAt(index);
        if (!vout.scriptPubKey.address)
            return true; //OP_RETURN
        const address = new BtcAddress_1.BtcAddress(vout.scriptPubKey.address);
        return address.addressToPkscript() == vout.scriptPubKey.hex;
    }
    get sourceAddresses() {
        if (this.type === "coinbase") {
            // Coinbase transactions mint coins
            return [undefined];
        }
        // Regular transactions
        return this.data.vin.map((vin) => {
            if (vin.prevout) {
                return vin.prevout.scriptPubKey.address;
            }
            else {
                // TODO: we don't have info about the source addresses of this transaction, return undefined for now
                // TODO: we can extract them from additional data (old way)
                return undefined;
            }
        });
    }
    get fee() {
        const error = new errors_1.mccError(errors_1.mccErrorCode.InvalidResponse, Error("fee can't be calculated, the transaction data does not contain the prevouts field (verbosity 2)"));
        if (this.type === "coinbase") {
            // Coinbase transactions mint coins
            return (0, utils_1.toBN)(0);
        }
        let inSum = (0, utils_1.toBN)(0);
        if (this.hasPrevouts) {
            inSum = this.data.vin.reduce(this.reducerFunctionPrevouts, (0, utils_1.toBN)(0));
        }
        else {
            throw error;
        }
        const outSum = this.data.vout.reduce(this.reducerFunctionVouts, (0, utils_1.toBN)(0));
        return inSum.sub(outSum);
    }
    get spentAmounts() {
        if (this.type === "coinbase") {
            // Coinbase transactions mint coins
            return [
                {
                    amount: (0, utils_1.toBN)(0),
                },
            ];
        }
        if (this.hasPrevouts) {
            return this.data.vin.map((vin) => {
                var _a;
                const address = vin.prevout && vin.prevout.scriptPubKey.address ? vin.prevout.scriptPubKey.address : undefined;
                const value = this.toBnValue(((_a = vin === null || vin === void 0 ? void 0 : vin.prevout) === null || _a === void 0 ? void 0 : _a.value) || 0);
                const addressAmount = {
                    address: address,
                    amount: value,
                    utxo: (vin === null || vin === void 0 ? void 0 : vin.vout) || 0,
                };
                return addressAmount;
            });
        }
        else {
            const error = new errors_1.mccError(errors_1.mccErrorCode.InvalidResponse, Error("fee can't be calculated, the transaction data does not contain the prevouts field (verbosity 2)"));
            throw error;
        }
    }
    get type() {
        if (this.data.vin.length === 0 || this.data.vin[0].coinbase) {
            return "coinbase";
        }
        if (this.hasPrevouts)
            return "full_payment";
        return "payment";
    }
    // BTC transaction specific helper methods
    get hasPrevouts() {
        return this.data.vin.every((vin) => {
            return vin.prevout !== undefined || vin.coinbase !== undefined;
        });
    }
}
exports.BtcTransaction = BtcTransaction;
//# sourceMappingURL=BtcTransaction.js.map